from dataclasses import asdict

from torch_sdk.errors import TorchSdkException
from torch_sdk.models.profile import ProfilingType, JobType
from torch_sdk.models.assetType import AssetType
from torch_sdk.models.create_asset import RelationType, CreateAssetRelation, AssetMetadata
from torch_sdk.models.datasource import DataSource, DatasourceSourceType


class AssetRelation:

    def __init__(self,
                 fromAssetId: str,
                 toAssetId: str,
                 currentSnapshot: str,
                 snapshots,
                 relation: RelationType,
                 hasSubRelations: bool = False,
                 id: int = None,
                 isDeleted: bool = None,
                 metadata=None,
                 **kwargs):
        """
            Description:
                Asset relation class
        :param fromAssetId: source asset id
        :param toAssetId: sink asset id
        :param currentSnapshot: current version of the asset relation
        :param snapshots: version list
        :param id: asset relation id
        :param isDeleted: isDeleted or not in newer version
        :param relation: (RelationType) relation type
        :param metadata: (List[AssetMetadata]) metadata of the asset relation
        :param hasSubRelations: (bool) has sub relations or not
        """
        self.fromAssetId = fromAssetId
        self.toAssetId = toAssetId
        self.currentSnapshot = currentSnapshot
        self.snapshots = snapshots
        self.id = id
        self.isDeleted = isDeleted
        self.relation = relation
        self.metadata = metadata
        self.hasSubRelations = hasSubRelations

    def __eq__(self, other):
        return self.toAssetId == other.toAssetUUID and self.fromAssetId == other.fromAssetUUID

    def __repr__(self):
        return f"AssetRelation({self.__dict__})"


class ChildType:

    def __init__(self, canProfile, canSample, id, name):
        self.canProfile = canProfile
        self.canSample = canSample
        self.id = id
        self.name = name

    def __eq__(self, other):
        return self.id == other.id

    def __repr__(self):
        return f"ChildType({self.__dict__})"


class ChildAsset:
    def __init__(self, assetId, isCustom, name, parentId=None, alias=None, childType=None):
        self.assetId = assetId
        self.isCustom = isCustom
        self.name = name
        self.parentId = parentId
        self.alias = alias
        if isinstance(childType, dict):
            self.datasource = ChildType(**childType)
        else:
            self.datasource = childType

    def __eq__(self, other):
        return self.assetId == other.assetId

    def __repr__(self):
        return f"ChildAsset({self.__dict__})"


class Metadata:

    def __init__(self, assetId, createdAt, updatedAt, currentSnapshot, id, items, metaDataHash, snapshots=None):
        self.assetId = assetId
        self.createdAt = createdAt
        self.updatedAt = updatedAt
        self.currentSnapshot = currentSnapshot
        self.id = id
        self.metaDataHash = metaDataHash
        if isinstance(items, list):
            items_list = list(items)
            asset_metadata = []
            for i in items_list:
                asset_mt = AssetMetadata(**i)
                asset_metadata.append(asset_mt)
            self.items = asset_metadata
        else:
            self.items = items
        if isinstance(snapshots, list):
            self.snapshots = list(snapshots)
        else:
            self.snapshots = snapshots

    def __eq__(self, other):
        return self.id == other.id

    def __repr__(self):
        return f"Metadata({self.__dict__})"


class Asset:

    def __init__(self,
                 alias=None,
                 assembly=None,
                 assetType: AssetType = None,
                 createdAt=None,
                 currentSnapshot=None,
                 description=None,
                 id=None,
                 isCustom=None,
                 isDeleted=None,
                 name=None,
                 parentId=None,
                 snapshots=None,
                 sourceType=None,
                 uid=None,
                 updatedAt=None,
                 client=None,
                 **kwargs
                 ):
        """
            Description:
                Asset class
        :param alias: alias of the asset
        :param assembly: (Datasource) data source details
        :param assetType: type of the asset
        :param createdAt: creation time of the asset
        :param currentSnapshot: current version of the asset
        :param description: desc of the asset
        :param id: asset id
        :param isCustom: is custom asset or not
        :param isDeleted: is deleted or not in current version of the datasource
        :param name: name of the asset
        :param parentId: parent id of the asset
        :param snapshots: version list
        :param sourceType: source type of the asset's datasource
        :param uid: uid of the asset
        :param updatedAt: updated time of the asset
        """
        self.alias = alias
        self.createdAt = createdAt
        self.currentSnapshot = currentSnapshot
        self.description = description
        self.id = id
        self.isCustom = isCustom
        self.isDeleted = isDeleted
        self.name = name
        self.parentId = parentId
        self.snapshots = snapshots
        self.uid = uid
        self.updatedAt = updatedAt
        if isinstance(assembly, dict):
            self.datasource = DataSource(**assembly)
        else:
            self.datasource = assembly
        if isinstance(assetType, dict):
            self.assetType = AssetType(**assetType)
        else:
            self.assetType = assetType
        if isinstance(sourceType, dict):
            self.sourceType = DatasourceSourceType(**sourceType)
        else:
            self.sourceType = sourceType

        # self.datasource = assembly
        # self.sourceType = sourceType

        self.client = client

    def __repr__(self):
        return f"Asset({self.__dict__})"

    # convert asset relation to dict type
    def _convert_asset_relation_to_dict(self, asset_relation: CreateAssetRelation):
        """
        Description:
            Convert CreateAssetRelation class instance to dict type
        :param assetRelation: CreateAssetRelation class instance
        :return: dict form of CreateAssetRelation class instance
        """
        payload = asset_relation.__dict__
        payload['relationType'] = asset_relation.relationType.name
        asset_relation_payload = {'data': payload}
        return asset_relation_payload

    # create relation between asset
    def create_asset_relation(self, to_asset_uuid: str, relation_type: RelationType,
                              snapshots=None):
        """
        Description:
            used to create relation between any 2 existing assets
        :param snapshots:
        :param relation_type:
        :param to_asset_uuid:
        :return: created AssetRelation class instance
        """
        if snapshots is None:
            snapshots = [self.currentSnapshot]
        asset_relation = CreateAssetRelation(
            fromAssetUUID=self.uid,
            assemblyId=self.datasource.id,
            toAssetUUID=to_asset_uuid,
            relationType=relation_type,
            currentSnapshot=self.currentSnapshot,
            snapshots=snapshots
        )
        payload = self._convert_asset_relation_to_dict(asset_relation)
        return self.client.create_asset_relation(payload)

    def get_related_assets(self):
        return self.client.get_related_assets(self.id)

    def get_child_assets(self):
        return self.client.get_child_assets(self.id)

    def start_watch(self):
        return self.client.start_watch(self.id)

    def stop_watch(self):
        return self.client.stop_watch(self.id)

    def get_asset_tags(self):
        return self.client.get_asset_tags(self.id)

    def get_asset_activity(self):
        return self.client.get_asset_activity(self.id)

    def get_asset_comment(self):
        return self.client.get_asset_comment(self.id)

    def add_asset_tag(self, tag):
        payload = {
            "assetTag": {
                "name": tag,
                "assetId": self.id,
                "autoTagged": False
            }
        }
        return self.client.add_asset_tag(self.id, payload)

    def remove_asset_tag(self, tag):
        tags = self.client.get_asset_tags(self.id)
        tag_id = None
        for tg in tags:
            if tg.name == tag:
                tag_id = tg.tagId
        if tag_id is None:
            raise TorchSdkException('Tag not found for the asset. Tag doesnt exists')
        return self.client.remove_asset_tag(self.id, tag_id)

    def add_asset_labels(self, labels: []):
        """
        Description:
            add asset labels.
            To give alias add 'alias:alias_name'.
            label key and value should be simple string. It should not contain any special characters in it.
        :param labels: list of strings. format : 'key:value'. Ex. labels : ['key1:value1', 'key2:value2']
        :return:
        """
        lbls = []
        for label in labels:
            count = label.count(':')
            if count != 1:
                raise TorchSdkException(
                    f'Invalid input label string {label}. label key and value should be simple string. It should not '
                    f'contain any special characters in it.')
            lbl = label.split(':')
            label_dict = {
                'key': lbl[0],
                'value': lbl[1]
            }
            lbls.append(label_dict)
        payload = {
            'labels': lbls
        }
        return self.client.add_asset_labels(self.id, payload)

    def get_asset_labels(self):
        return self.client.get_asset_labels(self.id)

    def update_asset_annotation(self, annotation: str):
        payload = {
            'annotation': annotation
        }
        return self.client.update_asset_annotation(self.id, payload)

    def get_asset_metadata(self):
        return self.client.get_asset_metadata(self.id)

    def update_asset_metadata(self, asset_metadata):
        """
        Description:
            Update asset metadata for a latest snapshot version
        :param asset_metadata: list of AssetMetadata Object
        :return:
        """
        metadata = []
        for md in asset_metadata:
            metadata.append(asdict(md))

        metadata_payload = {
            'data': {
                'items': metadata,
                'currentSnapshot': self.currentSnapshot,
                'snapshots': self.snapshots
            }
        }
        return self.client.update_asset_metadata(self.id, metadata_payload)

    def profile_asset(self, profiling_type: ProfilingType, job_type: JobType = JobType.PROFILE):
        payload = {
            'data': {
                'profilingType': profiling_type.value,
                'jobType': job_type.value
            }
        }
        return self.client.profile_asset(self.id, payload)

    def get_profile_status(self):
        return self.client.get_profile_status(self.id)

    def auto_tag_asset(self):
        payload = {
            'data': {
                'profilingType': ProfilingType.FULL.value,
            }
        }
        return self.client.auto_tag_asset(self.id, payload)

    def sample_data(self):
        return self.client.sample_data(self.id)

    def get_profile_request_details(self, req_id):
        self.client.get_profile_request_details(asset_id=self.id, req_id=req_id)


from enum import Enum


class ConnectionType:

    def __init__(self, id, type):
        self.id = id
        self.type = type

    def __eq__(self, other):
        return self.id == other.id

    def __repr__(self):
        return f"ConnectionType({self.__dict__})"


class AnalysisPipeline:

    def __init__(self, id, name, accessKey, createdAt, updatedAt, url, externalUrl, description, hbaseEnabled,
                 hdfsEnabled,
                 hiveEnabled, **kwargs):
        self.name = name
        self.accessKey = accessKey
        self.createdAt = createdAt
        self.updatedAt = updatedAt
        self.url = url
        self.id = id
        self.externalUrl = externalUrl
        self.description = description
        self.hbaseEnabled = hbaseEnabled
        self.hdfsEnabled = hdfsEnabled
        self.hiveEnable = hiveEnabled

    def __eq__(self, other):
        return self.name == other.name

    def __repr__(self):
        return f"AnalysisPipeline({self.__dict__})"


class ConfigProperty:

    def __init__(self, key, value, id=None, **kwargs):
        self.key = key
        self.value = value
        self.id = id

    def __eq__(self, other):
        return self.id == other.id

    def __repr__(self):
        return f"ConfigProperty({self.__dict__})"


class Connection:

    def __init__(self, name, connectionType, createdAt, updatedAt, analyticsPipeline, configuration, assemblyCount=0,
                 description=None,
                 id=None, **kwargs):
        self.name = name
        self.configuration = configuration
        self.connectionType = connectionType
        self.createdAt = createdAt
        self.updatedAt = updatedAt
        self.assemblyCount = assemblyCount
        self.description = description
        self.id = id
        if isinstance(connectionType, dict):
            self.connectionType = ConnectionType(**connectionType)
        else:
            self.connectionType = connectionType
        if isinstance(analyticsPipeline, dict):
            self.analyticsPipeline = AnalysisPipeline(**analyticsPipeline)
        else:
            self.analyticsPipeline = analyticsPipeline

    def __eq__(self, other):
        return self.name == other.name

    def __repr__(self):
        return f"Connection({self.__dict__})"


class CreateConnection:

    def __init__(self, name, connection_type_id, analytics_pipeline_id, properties, description=None):
        self.name = name
        self.connectionTypeId = connection_type_id
        self.analyticsPipelineId = analytics_pipeline_id
        self.properties = properties
        self.description = description

    def __eq__(self, other):
        return self.name == other.name

    def __repr__(self):
        return f"CreateConnection({self.__dict__})"


class ConnectionCheckStatus(Enum):
    SUCCESS = 1
    FAILURE = 2


class ConnectionCheck:

    def __init__(self, message, status, **kwargs):
        self.message = message
        self.status = status

    def __repr__(self):
        return f"ConnectionCheckResponse({self.__dict__})"



class Rule:

    def __init__(self, id, name, description, createdAt, updatedAt, backingAssets, thresholdLevel, archivalReason=None,
                 archived=False, enabled=False, schedule=None, notificationChannels=None, **kwargs):
        self.id = id
        self.name = name
        self.description = description
        self.createdAt = createdAt
        self.updatedAt = updatedAt
        self.backingAssets = backingAssets
        self.thresholdLevel = thresholdLevel
        self.archivalReason = archivalReason
        self.archived = archived
        self.enabled = enabled
        self.schedule = schedule
        self.notificationChannels = notificationChannels

    def __eq__(self, other):
        return self.id == other.id

    def __repr__(self):
        return f"Rule({self.__dict__})"


class RuleCancelResponse:

    def __init__(self, message, status, **kwargs):
        self.status = status
        self.message = message

    def __repr__(self):
        return f"Response({self.__dict__})"


class RuleExecution:

    def __init__(self, ruleId, ruleName, thresholdLevel, startedAt, resultStatus, executionStatus, executionMode
                 , ruleType=None, id=None, lastMarker=None, executionError=None, rightLastMarker=None,
                 leftLastMarker=None, finishedAt=None, **kwargs):
        self.ruleId = ruleId
        self.ruleName = ruleName
        self.thresholdLevel = thresholdLevel
        self.startedAt = startedAt
        self.resultStatus = resultStatus
        self.executionStatus = executionStatus
        self.executionMode = executionMode
        self.ruleType = ruleType
        self.executionError = executionError
        self.rightLastMarker = rightLastMarker
        self.lastMarker = lastMarker
        self.leftLastMarker = leftLastMarker
        self.id = id
        self.finishedAt = finishedAt

    def __eq__(self, other):
        return self.id == other.id

    def __repr__(self):
        return f"RuleExecution({self.__dict__})"


class RuleExecutionResult:

    def __init__(self, status, ruleExecutionType, qualityScore, description, failedRows, failureCount, rows, successCount,  **kwargs):
        self.status = status
        self.ruleExecutionType = ruleExecutionType
        self.description = description
        self.failedRows = failedRows
        self.failureCount = failureCount
        self.qualityScore = qualityScore
        self.rows = rows
        self.successCount = successCount

class ExecutionResult:

    def __init__(self, execution, items, meta = None, result= None, **kwargs):
        if isinstance(execution, dict):
            self.execution = RuleExecution(**execution)
        else:
            self.execution = execution
        if isinstance(result, dict):
            self.result = RuleExecutionResult(**result)
        else:
            self.result = result
        self.items = items
        self.meta = meta

    def __repr__(self):
        return f"ExecutionResult({self.__dict__})"



from torch_sdk.errors import TorchSdkException
from torch_sdk.models.pipeline import CreatePipeline
from torch_sdk.torch_http_client import TorchHttpClient
from torch_sdk.models.datasource import CreateDataSource


class TorchClient:
    """
            Description : Torch user client is used to send data to catalog server.

            :param url: (String) url of the catalog server
            :param timeout_ms: (Integer) timeout of the requests sending to catalog
            :param access_key: (String) Access key of API key. You can generate API key from torch UI's setting
            :param secret_key: (String) Secret key of API key.

            Ex.  TorchClient = TorchUserClient(url='https://torch.acceldata.local:5443', access_key='OY2VVIN2N6LJ', secret_key='da6bDBimQfXSMsyyhlPVJJfk7Zc2gs')
    """

    def __init__(self, url, timeout_ms=10000, access_key: str = None, secret_key: str = None):
        """
                Description : Torch user client is used to send data to catalog server.

                :param url: (String) url of the catalog server
                :param timeout_ms: (Integer) timeout of the requests sending to catalog
                :param access_key: (String) Access key of API key. You can generate API key from torch UI's setting
                :param secret_key: (String) Secret key of API key.

                Ex.  TorchClient = TorchUserClient(url='https://torch.acceldata.local:5443', access_key='OY2VVIN2N6LJ', secret_key='da6bDBimQfXSMsyyhlPVJJfk7Zc2gs')
        """

        if access_key is None and secret_key is None:
            raise Exception('Access key and secret key - required')
        self.client = TorchHttpClient(url=url, access_key=access_key, secret_key=secret_key, timeout_ms=timeout_ms)

    def create_pipeline(self, pipeline: CreatePipeline):
        """
        Description:
            To create pipeline in torch catalog service
        :param pipeline: (CreatePipeline) class instance of the pipeline to be created
        :return: (Pipeline) newly created pipeline class instance
        """
        if pipeline.uid is None or pipeline.name is None:
            raise Exception('To create a pipeline, pipeline uid/name is required')
        return self.client.create_pipeline(pipeline)

    def get_pipeline(self, uid: str):
        """
        Description:
            To get an existing pipeline from torch catalog
        :param uid: uid of the pipeline
        :return:(Pipeline) pipeline class instance
        """
        if uid is None:
            raise Exception('To get a pipeline, pipeline uid is required')
        return self.client.get_pipeline(uid)

    def create_datasource(self, datasource: CreateDataSource):
        """
        Description:
            To create datasource in torch catalog
        :param datasource: (CreateDataSource) class instance of the datasource to be created
        :return: (DataSource) newly created datasource
        """
        return self.client.create_datasource(datasource)

    def get_datasource(self, name: str):
        """
        Description:
            Find datasource by it's name in torch catalog
        :param name: name of the datasource given in torch
        :return: (DataSource) datasource
        """
        return self.client.get_datasource(name)

    def get_all_datasources(self):
        """
        Description:
            list all datasources in torch catalog
        :return: (DataSource) list of datasource
        """
        return self.client.get_all_datasources()

    def get_asset_types(self):
        """
        Description:
            get all asset types supported in torch catalog
        :return: list of asset types
        """
        return self.client.get_all_asset_types()

    def get_all_source_types(self):
        """
        Description:
            get all source types supported in torch catalog
        :return: list of all source type
        """
        return self.client.get_all_source_types()

    def get_property_templates(self):
        pass

    def get_connection_types(self):
        """
        Description:
            get all connection types supported in torch catalog
        :return: list of all connection types
        """
        return self.client.get_connection_types()

    def get_tags(self):
        return self.client.get_tags()

    def delete_tag(self, tag: str):
        return self.client.delete_tag(tag)

    def update_data_protection(self, tag: str, is_protected: bool):
        return self.client.update_data_protection(tag, is_protected)

    def discover(self):
        # add filters in discover api & more details in asset class
        return self.client.discover()

    def list_all_snapshots(self):
        return self.client.list_all_snapshots()

    def get_user(self, user_id=None, username=None):
        if user_id is None and username is None:
            raise TorchSdkException('Either provide uid or id to find an asset')
        if user_id is not None:
            return self.client.get_user_by_id(id=user_id)
        if username is not None:
            return self.client.get_user_by_name(username=username)

    def create_api_key(self):
        return self.client.create_api_keys()

    def get_api_keys(self):
        return self.client.get_api_keys()

    def delete_api_key(self, access_key: str):
        return self.client.delete_api_key(access_key)

    def get_notifications(self):
        return self.client.get_notifications()

    def get_incidents(self):
        return self.client.get_incidents()

    def list_all_analysis_pipelines(self):
        return self.client.list_all_analysis_pipelines()

    def get_analysis_pipeline(self, id: int):
        return self.client.get_analysis_pipeline(id)

    def get_all_auto_profile_configurations(self):
        return self.client.get_all_auto_profile_configurations()

    def get_asset(self, uid: str = None, id: int = None):
        """"
            Description:
                Find an asset of the datasource
            :param uid: (String) uid of the asset
            :param id: (Int) id of the asset in torch catalog
        """
        if uid is None and id is None:
            raise TorchSdkException('Either provide uid or id to find an asset')
        if uid is not None:
            return self.client.get_asset_by_uid(uid=uid)
        if id is not None:
            return self.client.get_asset_by_id(id=id)

    def list_all_classification_configurations(self):
        return self.client.list_all_classification_configurations()

    def create_classification_configuration(self, classification_config):
        return self.client.create_classification_configuration(classification_config)

    def update_classification_configuration(self, classification_config):
        return self.client.update_classification_configuration(classification_config)

    def get_classification_configuration(self, name: str):
        return self.client.get_classification_configuration(name)

    def delete_classification_configuration(self, name: str):
        return self.client.delete_classification_configuration(name)

    def list_setting_groups(self):
        return self.client.list_setting_groups()

    def find_settings_by_group(self, setting_group_name: str):
        return self.client.find_settings_by_group(setting_group_name)

    def get_setting(self, key: str):
        return self.client.get_setting(key)

    def update_setting_value(self, key: str, value: str):
        return self.client.update_setting_value(key, value)

    def reset_setting_value(self, key:str):
        return self.client.reset_setting_value(key)

    def list_all_connections(self):
        return self.client.list_all_connections()

    def create_connection(self, create_connection):
        return self.client.create_connection(create_connection)

    def update_connection(self, update_connection):
        if update_connection.name is None:
            raise TorchSdkException('Update connection should have name of the connection. (update_connection : '
                                    'CreateConnection class instance)')
        return self.client.update_connection(update_connection.name, update_connection)

    def delete_connection(self, id):
        return self.client.delete_connection(id)

    def check_connection(self, connection):
        return self.client.check_connection(connection)

    def get_connection(self, name):
        return self.client.get_connection(name)

    def execute_dq_rule(self, rule_id):
        return self.client.execute_dq_rule(rule_id)

    def execute_reconciliation_rule(self, rule_id):
        return self.client.execute_reconciliation_rule(rule_id)

    def get_dq_rule_execution_details(self, execution_id):
        return self.client.get_dq_rule_execution_details(execution_id)

    def get_reconciliation_rule_execution_details(self, execution_id):
        return self.client.get_reconciliation_rule_execution_details(execution_id)

    def cancel_rule_execution(self, execution_id):
        return self.client.cancel_rule_execution(execution_id)

    def enable_rule(self, rule_id):
        return self.client.enable_rule(rule_id)

    def disable_rule(self, rule_id):
        return self.client.disable_rule(rule_id)

    def get_reconciliation_rule_result(self, execution_id):
        return self.client.get_reconciliation_rule_result(execution_id)

    def get_dq_rule_result(self, execution_id):
        return self.client.get_dq_rule_result(execution_id)
